#include "HidInjector.h"
#include "hdmi_rpi.h"

#include <stdio.h>
#include <stdarg.h>
#include <cfgmgr32.h> // Requires WDK for Raspberry Pi(ARM architecture)
#include <initguid.h>
#include <math.h>

// {F990ABA9-C9C0-4CF8-A4A2-5B06D97F6871}
DEFINE_GUID(GUID_DEVCLASS_HIDINJECTOR, 0xf990aba9, 0xc9c0, 0x4cf8, 0xa4, 0xa2, 0x5b, 0x6, 0xd9, 0x7f, 0x68, 0x71);
// {59819B74-F102-469A-9009-3CAF35FD4686}
DEFINE_GUID(GUID_DEVINTERFACE_HIDINJECTOR, 0x59819b74, 0xf102, 0x469a, 0x90, 0x9, 0x3c, 0xaf, 0x35, 0xfd, 0x46, 0x86);
// {FFD216E4-A560-4676-8BD5-4F26A5BFF214}
DEFINE_GUID(GUID_BUS_HIDINJECTOR, 0xffd216e4, 0xa560, 0x4676, 0x8b, 0xd5, 0x4f, 0x26, 0xa5, 0xbf, 0xf2, 0x14);

int HidInjector::hcount = 0;
HANDLE HidInjector::handle = INVALID_HANDLE_VALUE;

int HidInjector::sc_width = 1920;
int HidInjector::sc_height = 1080;

HIDINJECTOR_INPUT_REPORT HidInjector::mouse_state = { 0 };
HIDINJECTOR_INPUT_REPORT HidInjector::key_state = { 0 };
HIDINJECTOR_INPUT_REPORT HidInjector::touch_state = { 0 };

BOOL HidInjector::mouseUsed = FALSE;
TouchData HidInjector::touches[TOUCH_MAX_FINGER] = { 0 };

HidInjector::HidInjector()
{
	mouse_state.ReportId = MOUSE_REPORT_ID;
	key_state.ReportId = KEYBOARD_REPORT_ID;
	touch_state.ReportId = TOUCH_REPORT_ID;
}

HidInjector::~HidInjector()
{
	Close();
}

void HidInjector::OutputMessage(LPCWSTR msg, ...)
{
#ifdef _DEBUG
	wchar_t fmsg[260];
	va_list argptr;
	va_start(argptr, msg);
	vswprintf_s(fmsg, 260, msg, argptr);
	va_end(argptr);

	OutputDebugString(fmsg);
#else
	va_list argptr;
	va_start(argptr, msg);
	vwprintf_s(msg, argptr);
	va_end(argptr);
#endif
}

BOOL HidInjector::FindMatchingDevice()
{
	BOOL res = FALSE;

	PWSTR deviceInterfaceList = NULL;
	GUID devInterface = GUID_DEVINTERFACE_HIDINJECTOR;
	ULONG deviceInterfaceListLength = 0;

	CONFIGRET cr;
	cr = CM_Get_Device_Interface_List_Size(
		&deviceInterfaceListLength, &devInterface,
		NULL, CM_GET_DEVICE_INTERFACE_LIST_PRESENT);
	if (cr != CR_SUCCESS) {
		OutputMessage(L"Error 0x%x retrieving device interface list size.\n", cr);
		goto EXIT;
	}

	if (deviceInterfaceListLength <= 1) {
		OutputMessage(L"Error: No active device interfaces found.\n Is the sample driver loaded?\n");
		goto EXIT;
	}

	deviceInterfaceList = new wchar_t[deviceInterfaceListLength];
	if (deviceInterfaceList == NULL) {
		OutputMessage(L"Error allocating memory for device interface list.\n");
		goto EXIT;
	}
	ZeroMemory(deviceInterfaceList, deviceInterfaceListLength * sizeof(WCHAR));

	cr = CM_Get_Device_Interface_List(&devInterface, NULL,
		deviceInterfaceList, deviceInterfaceListLength,
		CM_GET_DEVICE_INTERFACE_LIST_PRESENT);
	if (cr != CR_SUCCESS) {
		OutputMessage(L"Error 0x%x retrieving device interface list.\n", cr);
		goto EXIT;
	}

	OutputMessage(L"...looking for our HID device.\n");

	for (PWSTR ci = deviceInterfaceList; *ci; ci += wcslen(ci) + 1) {
		OutputMessage(L"Aquiring %S ", ci);
		HANDLE devHandle = CreateFile(ci,
			GENERIC_READ | GENERIC_WRITE,
			FILE_SHARE_READ | FILE_SHARE_WRITE,
			NULL, OPEN_EXISTING, 0, NULL);
		if (devHandle != INVALID_HANDLE_VALUE) {
			OutputMessage(L"is succeeded.");
			handle = devHandle;
			res = TRUE;
			break;
		}

		OutputMessage(L"is failed(%d)\n", GetLastError());
	}

EXIT:
	if (deviceInterfaceList) {
		delete[]deviceInterfaceList;
	}

	return res;
}

BOOL HidInjector::Open()
{
	if (hcount > 0) {
		hcount++;
		return TRUE;
	}

	BOOL success = FindMatchingDevice();
	if (!success) {
		handle = INVALID_HANDLE_VALUE;
		OutputMessage(L"Device is not found.");
	} else {
		// Get display resolution
#ifndef _MBM_BARE
		FILE *fp = NULL;
		if (fopen_s(&fp, "C:\\EFIESP\\config.txt", "r") == 0) {
			char line[260], *p;
			int hdmi_group = 0, hdmi_mode = 0, *v;
			while (fgets(line, 255, fp) != NULL && (hdmi_group == 0 || hdmi_mode == 0)) {
				if (p = strstr(line, "hdmi_group")) {
					v = &hdmi_group;
					p += 11;
				} else if (p = strstr(line, "hdmi_mode")) {
					v = &hdmi_mode;
					p += 10;
				} else {
					continue;
				}

				while (*p < '0' || *p > '9') p++;

				char *lp = line; int n = 0;
				while (*p >= '0' && *p <= '9') {
					*lp = *p;
					lp++; p++; n++;
				}
				*lp = '\0';

				*v = 0;
				for (int i = 0; i < n; i++) {
					*v = (line[i] - '0') * (int)pow(10, n - i - 1);
				}
			}

			if (hdmi_group == 2 && hdmi_mode <= HDMI_MAX_VALUE) {
				sc_width = HdmiDmtDefines[hdmi_mode][1];
				sc_height = HdmiDmtDefines[hdmi_mode][2];
			}
		}
#endif
	}
	return success;
}

void HidInjector::Close()
{
	hcount--;
	if(hcount <= 0){
		if (handle != INVALID_HANDLE_VALUE) {
			// Reset all inputs
			for (int i = 0; i < 4; i++) KeyUp(i);
			for (int i = 0; i < TOUCH_MAX_FINGER; i++) TouchUp(i);
			if(mouseUsed == TRUE) MouseUp(MOUSE_BUTTON_1 | MOUSE_BUTTON_2 | MOUSE_BUTTON_3);

			CloseHandle(handle);
			handle = INVALID_HANDLE_VALUE;
		}
		hcount = 0;
	}
}

BOOL HidInjector::SendReport(HIDINJECTOR_INPUT_REPORT *report) 
{
	if (handle == INVALID_HANDLE_VALUE) return FALSE;

	DWORD written = 0;
	return WriteFile(handle, report, sizeof(HIDINJECTOR_INPUT_REPORT), &written, NULL);
}

BOOL HidInjector::SendKey(int pos, UsbKeys uk)
{
	switch (pos) {
	case 1: key_state.Report.KeyReport.Key2 = uk; break;
	case 2: key_state.Report.KeyReport.Key3 = uk; break;
	case 3: key_state.Report.KeyReport.Key4 = uk; break;
	default: key_state.Report.KeyReport.Key1 = uk; break;
	}
	return SendReport(&key_state);
}

BOOL HidInjector::KeyDown(int pos, UsbKeys uk)
{
	return SendKey(pos, uk);
}

BOOL HidInjector::KeyUp(int pos)
{
	return SendKey(pos, UsbKeys::None);
}

BOOL HidInjector::SendMouse(int x, int y, UCHAR buttons)
{
	mouseUsed = TRUE;
	mouse_state.Report.MouseReport.Buttons = buttons;
	return SendReport(&mouse_state);
}

BOOL HidInjector::MouseMove(int x, int y)
{
	if (x < 0) x = 0; else if (x >= sc_width) x = sc_width - 1;
	if (y < 0) y = 0; else if (y >= sc_height) y = sc_height - 1;

	mouse_state.Report.MouseReport.AbsoluteX = x * 32767 / sc_width;
	mouse_state.Report.MouseReport.AbsoluteY = y * 32767 / sc_height;
	return SendReport(&mouse_state);
}

BOOL HidInjector::MouseDown(UCHAR pos)
{
	mouse_state.Report.MouseReport.Buttons |= pos;
	return SendReport(&mouse_state);
}

BOOL HidInjector::MouseUp(UCHAR pos)
{
	mouse_state.Report.MouseReport.Buttons &= ~pos;
	return SendReport(&mouse_state);
}

BOOL HidInjector::SendTouch(int pos)
{
	TouchData *t = &touches[pos];
	touch_state.Report.TouchReport.ContactCount = 1;
	touch_state.Report.TouchReport.ContactIndentifier = pos;
	touch_state.Report.TouchReport.AbsoluteX = (USHORT)t->x;
	touch_state.Report.TouchReport.AbsoluteY = (USHORT)t->y;
	touch_state.Report.TouchReport.Flags = (t->down == TRUE) ? TOUCH_TIP_SWITCH : 0;

	return SendReport(&touch_state);
}

BOOL HidInjector::TouchDown(UCHAR pos)
{
	if (pos < 0 || pos >= TOUCH_MAX_FINGER) return FALSE;

	TouchData *t = &touches[pos];
	t->down = TRUE;
	return SendTouch(pos);
}

BOOL HidInjector::TouchMove(UCHAR pos, int x, int y)
{
	if (pos < 0 || pos >= TOUCH_MAX_FINGER) return FALSE;

	if (x < 0) x = 0; else if (x >= sc_width) x = sc_width - 1;
	if (y < 0) y = 0; else if (y >= sc_height) y = sc_height - 1;

	TouchData *t = &touches[pos];
	t->x = x * TOUCH_PHYSICAL_MAX / sc_width;
	t->y = y * TOUCH_PHYSICAL_MAX / sc_height;
	return SendTouch(pos);
}

BOOL HidInjector::TouchUp(UCHAR pos)
{
	if (pos < 0 || pos >= TOUCH_MAX_FINGER) return FALSE;

	TouchData *t = &touches[pos];
	t->down = FALSE;
	return SendTouch(pos);
}
