// Copyright (c) Microsoft Open Technologies, Inc.  All rights reserved.  
// Licensed under the BSD 2-Clause License.  
// See License.txt in the project root for license information.

#include "common.h"

#include <concrt.h>
#include "ErrorCodes.h"
#include "DmapSupport.h"

/**
Get the base address of a memory mapped controller in the SOC.  Exclusive, non-shared
access to the controller is requested.
\param[in] deviceName The name of the PCI device used to map the controller in question.
\param[out] handle Handle opened to the device specified by deviceName.
\param[out] baseAddress Base address of the controller in question.
\return HRESULT success or error code.
*/
HRESULT GetControllerBaseAddress(PWCHAR deviceName, HANDLE & handle, PVOID & baseAddress)
{
	return GetControllerBaseAddress(deviceName, handle, baseAddress, 0);
}

/**
Get the base address of a memory mapped controller in the SOC.
\param[in] deviceName The name of the PCI device used to map the controller in question.
\param[out] handle Handle opened to the device specified by deviceName (zero for UWP build).
\param[out] baseAddress Base address of the controller in question.
\param[in] shareMode Sharing specifier as specified to CreateFile().
\return HRESULT success or error code.
*/
HRESULT GetControllerBaseAddress(PWCHAR deviceName, HANDLE & handle, PVOID & baseAddress, DWORD shareMode)
{
	void* controllerAddress = 0;
	HRESULT hr = S_OK;

	DMAP_MAPMEMORY_OUTPUT_BUFFER buf = { 0 };
	DWORD bytesReturned = 0;

	hr = OpenControllerDevice(deviceName, handle, shareMode);

	if (SUCCEEDED(hr) && (baseAddress == nullptr)){
		// Retrieve the base address of controller registers.
		if (!DeviceIoControl(handle, IOCTL_DMAP_MAPMEMORY, nullptr,
		0, &buf, sizeof(buf), &bytesReturned, nullptr)) {
			hr = HRESULT_FROM_WIN32(GetLastError());
			CloseHandle(handle);
			handle = INVALID_HANDLE_VALUE;
			baseAddress = nullptr;
		} else {
			// Pass back the base address of the controller registers to the caller.
			baseAddress = buf.Address;
		}
	}

	return hr;
}

/**
Close an open controller device using a method appropriate to how it was opened.
\param[inout] handle Handle to the open controller device.
*/
void DmapCloseController(HANDLE & handle)
{
	if (handle != INVALID_HANDLE_VALUE){
		CloseHandle(handle);
		handle = INVALID_HANDLE_VALUE;
	}
}

/**
Open a controller device in the SOC.  This is used for both memory mapped and
IO mapped controllers.
\param[in] deviceName The name of the PCI device used to map the controller in question.
\param[out] handle Handle opened to the device specified by deviceName.
\param[in] shareMode Sharing specifier as specified to CreateFile().
\return HRESULT success or error code.
*/
HRESULT OpenControllerDevice(PWCHAR deviceName, HANDLE & handle, DWORD shareMode)
{
	HRESULT hr = S_OK;

	if (handle == INVALID_HANDLE_VALUE){
		// Open the specified device.
		handle = CreateFile(deviceName, GENERIC_READ | GENERIC_WRITE,
			shareMode, nullptr, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

		if (handle == INVALID_HANDLE_VALUE){
			hr = HRESULT_FROM_WIN32(GetLastError());
			if (SUCCEEDED(hr))
			{
				hr = E_HANDLE;
			}
		}
	}

	return hr;
}
