// Copyright (c) Microsoft Open Technologies, Inc.  All rights reserved.  
// Licensed under the BSD 2-Clause License.  
// See License.txt in the project root for license information.


#include "common.h"

#include "ErrorCodes.h"
#include "GpioController.h"
#include "DmapSupport.h"

#if defined(_M_IX86) || defined(_M_X64)

// 
// Global extern exports
//

/// The global object used to interact with the BayTrail Fabric GPIO hardware.
BtFabricGpioControllerClass g_btFabricGpio;

/**
\return HRESULT success or error code.
*/
HRESULT BtFabricGpioControllerClass::_mapS0Controller()
{
	HRESULT hr = S_OK;
	PVOID baseAddress = nullptr;

	hr = GetControllerBaseAddress(
		mbmGpioS0DeviceName,
		m_hS0Controller,
		baseAddress,
		FILE_SHARE_READ | FILE_SHARE_WRITE);

	if (SUCCEEDED(hr))
	{
		m_s0Controller = (PGPIO_PAD)baseAddress;
	}

	return hr;
}
#endif // defined(_M_IX86) || defined(_M_X64)

#if defined(_M_IX86) || defined(_M_X64)
/**
\return HRESULT success or error code.
*/
HRESULT BtFabricGpioControllerClass::_mapS5Controller()
{
	HRESULT hr = S_OK;
	
	PVOID baseAddress = nullptr;

	hr = GetControllerBaseAddress(
		mbmGpioS5DeviceName,
		m_hS5Controller,
		baseAddress,
		FILE_SHARE_READ | FILE_SHARE_WRITE);

	if (SUCCEEDED(hr))
	{
		m_s5Controller = (PGPIO_PAD)baseAddress;
	}

	return hr;
}
#endif // defined(_M_IX86) || defined(_M_X64)

#if defined(_M_IX86) || defined(_M_X64)
/**
This method assumes the caller has checked the input parameters.
\param[in] gpioNo The S0 GPIO number of the pad to set. Range: 0-127.
\param[in] state State to set the pad to. 0 - LOW, 1 - HIGH.
\return HRESULT error or success code.
*/
HRESULT BtFabricGpioControllerClass::setS0PinState(ULONG gpioNo, ULONG state)
{
	HRESULT hr = mapS0IfNeeded();
	if (SUCCEEDED(hr)) {
		_PAD_VAL padVal;
		padVal.ALL_BITS = m_s0Controller[gpioNo].PAD_VAL.ALL_BITS;
		if (state == 0) padVal.PAD_VAL = 0;
		else padVal.PAD_VAL = 1;
		m_s0Controller[gpioNo].PAD_VAL.ALL_BITS = padVal.ALL_BITS;
	}

	return hr;
}
#endif // defined(_M_IX86) || defined(_M_X64)

#if defined(_M_IX86) || defined(_M_X64)
/**
This method assumes the caller has checked the input parameters.
\param[in] gpioNo The S5 GPIO number of the pad to set. Range: 0-59.
\param[in] state State to set the pad to. 0 - LOW, 1 - HIGH.
\return HRESULT error or success code.
*/
HRESULT BtFabricGpioControllerClass::setS5PinState(ULONG gpioNo, ULONG state)
{
	HRESULT hr = mapS5IfNeeded();
	if (SUCCEEDED(hr)) {
		_PAD_VAL padVal;
		padVal.ALL_BITS = m_s5Controller[gpioNo].PAD_VAL.ALL_BITS;
		if (state == 0) padVal.PAD_VAL = 0;
		else padVal.PAD_VAL = 1;
		m_s5Controller[gpioNo].PAD_VAL.ALL_BITS = padVal.ALL_BITS;
	}

	return hr;
}
#endif // defined(_M_IX86) || defined(_M_X64)

#if defined(_M_IX86) || defined(_M_X64)
/**
This method assumes the caller has checked the input parameters.
\param[in] gpioNo The S0 GPIO number of the pad to read. Range: 0-127.
\param[out] state Set to the state of the input bit.  0 - LOW, 1 - HIGH.
\return HRESULT error or success code.
*/
HRESULT BtFabricGpioControllerClass::getS0PinState(ULONG gpioNo, ULONG & state)
{
	HRESULT hr = mapS0IfNeeded();

	if (SUCCEEDED(hr)) {
		_PAD_VAL padVal;
		padVal.ALL_BITS = m_s0Controller[gpioNo].PAD_VAL.ALL_BITS;
		state = padVal.PAD_VAL;
	}

	return hr;
}
#endif // defined(_M_IX86) || defined(_M_X64)

#if defined(_M_IX86) || defined(_M_X64)
/**
This method assumes the caller has checked the input parameters.
\param[in] gpioNo The S5 GPIO number of the pad to read. Range: 0-59.
\param[out] state Set to the state of the input bit.  0 - LOW, 1 - HIGH.
\return HRESULT error or success code.
*/
HRESULT BtFabricGpioControllerClass::getS5PinState(ULONG gpioNo, ULONG & state)
{
	HRESULT hr = mapS5IfNeeded();
	if (SUCCEEDED(hr)) {
		_PAD_VAL padVal;
		padVal.ALL_BITS = m_s5Controller[gpioNo].PAD_VAL.ALL_BITS;
		state = padVal.PAD_VAL;
	}

	return hr;
}
#endif // defined(_M_IX86) || defined(_M_X64)

#if defined(_M_IX86) || defined(_M_X64)
/**
This method assumes the caller has checked the input parameters.
\param[in] gpioNo The S0 GPIO number of the pad to configure. Range: 0-127.
\param[in] mode The mode to set for the bit.  Range: DIRECTION_IN or DIRECTION_OUT
\return HRESULT error or success code.
*/
HRESULT BtFabricGpioControllerClass::setS0PinDirection(ULONG gpioNo, ULONG mode)
{
	HRESULT hr = S_OK;

	hr = mapS0IfNeeded();

	if (SUCCEEDED(hr)) {
		switch (mode) {
		case DIRECTION_IN: _setS0PinInput(gpioNo); break;
		case DIRECTION_OUT: _setS0PinOutput(gpioNo); break;
		}
	}

	return hr;
}
#endif // defined(_M_IX86) || defined(_M_X64)

#if defined(_M_IX86) || defined(_M_X64)
/**
This method assumes the caller has checked the input parameters.
\param[in] gpioNo The S5 GPIO number of the pad to configure. Range: 0-59.
\param[in] mode The mode to set for the bit.  Range: DIRECTION_IN or DIRECTION_OUT
\return HRESULT error or success code.
*/
inline HRESULT BtFabricGpioControllerClass::setS5PinDirection(ULONG gpioNo, ULONG mode)
{
	HRESULT hr = S_OK;

	hr = mapS5IfNeeded();

	if (SUCCEEDED(hr)) {
		switch (mode) {
		case DIRECTION_IN: _setS5PinInput(gpioNo); break;
		case DIRECTION_OUT: _setS5PinOutput(gpioNo); break;
		}
	}

	return hr;
}
#endif // defined(_M_IX86) || defined(_M_X64)

#if defined(_M_IX86) || defined(_M_X64)
/**
This method assumes the caller has checked the input parameters.
\param[in] gpioNo The S0 GPIO number of the pad to configure.  Range: 0-127.
\param[in] function The function to set for the pin.  Range: 0-7 (only 0-1 used here).
\return HRESULT error or success code.
*/
HRESULT BtFabricGpioControllerClass::setS0PinFunction(ULONG gpioNo, ULONG function)
{
	HRESULT hr = S_OK;

	hr = mapS0IfNeeded();

	if (SUCCEEDED(hr)) {
		_PCONF0 padConfig;
		padConfig.ALL_BITS = m_s0Controller[gpioNo].PCONF0.ALL_BITS;
		padConfig.FUNC_PIN_MUX = function;
		m_s0Controller[gpioNo].PCONF0.ALL_BITS = padConfig.ALL_BITS;
	}

	return hr;
}
#endif // defined(_M_IX86) || defined(_M_X64)

#if defined(_M_IX86) || defined(_M_X64)
/**
This method assumes the caller has checked the input parameters.
\param[in] gpioNo The S5 GPIO number of the pad to configure.  Range: 0-59.
\param[in] function The function to set for the pin.  Range: 0-7 (only 0-1 used here).
\return HRESULT error or success code.
*/
HRESULT BtFabricGpioControllerClass::setS5PinFunction(ULONG gpioNo, ULONG function)
{
	HRESULT hr = S_OK;

	hr = mapS5IfNeeded();

	if (SUCCEEDED(hr)) {
		_PCONF0 padConfig;
		padConfig.ALL_BITS = m_s5Controller[gpioNo].PCONF0.ALL_BITS;
		padConfig.FUNC_PIN_MUX = function;
		m_s5Controller[gpioNo].PCONF0.ALL_BITS = padConfig.ALL_BITS;
	}

	return hr;
}
#endif // defined(_M_IX86) || defined(_M_X64)

#if defined(_M_IX86) || defined(_M_X64)
/**
This routine disables the output latch for the pad, disables pin output and
enables pin input.
\param[in] gpioNo The S0 GPIO number of the pad to configure. Range: 0-127.
*/
void BtFabricGpioControllerClass::_setS0PinInput(ULONG gpioNo)
{
	_PCONF0 padConfig;
	padConfig.ALL_BITS = m_s0Controller[gpioNo].PCONF0.ALL_BITS;
	padConfig.BYPASS_FLOP = 1;		 // Disable flop
	padConfig.PULL_ASSIGN = 0;		 // Disable pull-up
	m_s0Controller[gpioNo].PCONF0.ALL_BITS = padConfig.ALL_BITS;

	_PAD_VAL padVal;
	padVal.ALL_BITS = m_s0Controller[gpioNo].PAD_VAL.ALL_BITS;
	padVal.IINENB = 0;				 // Enable pad for input
	padVal.IOUTENB = 1;				// Disable pad for output
	m_s0Controller[gpioNo].PAD_VAL.ALL_BITS = padVal.ALL_BITS;
}
#endif // defined(_M_IX86) || defined(_M_X64)

#if defined(_M_IX86) || defined(_M_X64)
/**
This routine disables the output latch for the pad, disables pin output and
enables pin input.
\param[in] gpioNo The S5 GPIO number of the pad to configure. Range: 0-127.
*/
void BtFabricGpioControllerClass::_setS5PinInput(ULONG gpioNo)
{
	_PCONF0 padConfig;
	padConfig.ALL_BITS = m_s5Controller[gpioNo].PCONF0.ALL_BITS;
	padConfig.BYPASS_FLOP = 1;		 // Disable flop
	padConfig.PULL_ASSIGN = 0;		 // No pull resistor
	m_s5Controller[gpioNo].PCONF0.ALL_BITS = padConfig.ALL_BITS;

	_PAD_VAL padVal;
	padVal.ALL_BITS = m_s5Controller[gpioNo].PAD_VAL.ALL_BITS;
	padVal.IINENB = 0;				 // Enable pad for input
	padVal.IOUTENB = 1;				// Disable pad for output
	m_s5Controller[gpioNo].PAD_VAL.ALL_BITS = padVal.ALL_BITS;
}
#endif // defined(_M_IX86) || defined(_M_X64)

#if defined(_M_IX86) || defined(_M_X64)
/**
This routine enables the output latch for the pad, disables pull-ups on
the pad, disables pin input and enables pin output.
\param[in] gpioNo The S0 GPIO number of the pad to configure. Range: 0-127.
*/
void BtFabricGpioControllerClass::_setS0PinOutput(ULONG gpioNo)
{
	_PCONF0 padConfig;
	padConfig.ALL_BITS = m_s0Controller[gpioNo].PCONF0.ALL_BITS;
	padConfig.BYPASS_FLOP = 0;		 // Enable flop
	padConfig.PULL_ASSIGN = 0;		 // No pull resistor
	padConfig.FUNC_PIN_MUX = 0;		// Mux function 0 (GPIO)
	m_s0Controller[gpioNo].PCONF0.ALL_BITS = padConfig.ALL_BITS;

	_PAD_VAL padVal;
	padVal.ALL_BITS = m_s0Controller[gpioNo].PAD_VAL.ALL_BITS;
	padVal.IOUTENB = 0;				// Enable pad for output
	padVal.IINENB = 1;				 // Disable pad for input
	m_s0Controller[gpioNo].PAD_VAL.ALL_BITS = padVal.ALL_BITS;
}
#endif // defined(_M_IX86) || defined(_M_X64)

#if defined(_M_IX86) || defined(_M_X64)
/**
This routine enables the output latch for the pad, disables pull-ups on
the pad, disables pin input and enables pin output.
\param[in] gpioNo The S5 GPIO number of the pad to configure. Range: 0-127.
*/
void BtFabricGpioControllerClass::_setS5PinOutput(ULONG gpioNo)
{
	_PCONF0 padConfig;
	padConfig.ALL_BITS = m_s5Controller[gpioNo].PCONF0.ALL_BITS;
	padConfig.BYPASS_FLOP = 0;		 // Enable flop
	padConfig.PULL_ASSIGN = 0;		 // No pull resistor
	m_s5Controller[gpioNo].PCONF0.ALL_BITS = padConfig.ALL_BITS;

	_PAD_VAL padVal;
	padVal.ALL_BITS = m_s5Controller[gpioNo].PAD_VAL.ALL_BITS;
	padVal.IOUTENB = 0;				// Enable pad for output
	padVal.IINENB = 1;				 // Disable pad for input
	m_s5Controller[gpioNo].PAD_VAL.ALL_BITS = padVal.ALL_BITS;
}
#endif // defined(_M_IX86) || defined(_M_X64)

#if defined(_M_IX86) || defined(_M_X64)
/// Method to detach an interrupt for an S0 GPIO port bit.
HRESULT BtFabricGpioControllerClass::detachS0Interrupt(ULONG intNo)
{
	HRESULT hr = S_OK;

	hr = mapS0IfNeeded();

	// Tell the driver to detach the interrupt.
	if (SUCCEEDED(hr))
	{
		hr = m_gpioInterrupts.detachInterrupt(intNo, m_hS0Controller);
	}

	return hr;
}
#endif // defined(_M_IX86) || defined(_M_X64)

#if defined(_M_IX86) || defined(_M_X64)
/// Method to detach an interrupt for an S5 GPIO port bit.
HRESULT BtFabricGpioControllerClass::detachS5Interrupt(ULONG intNo)
{
	HRESULT hr = S_OK;

	hr = mapS5IfNeeded();

	// Tell the driver to detach the interrupt.
	if (SUCCEEDED(hr))
	{
		hr = m_gpioInterrupts.detachInterrupt(intNo, m_hS5Controller);
	}

	return hr;
}
#endif // defined(_M_IX86) || defined(_M_X64)


#if defined(_M_ARM)
BcmGpioControllerClass g_bcmGpio;

/**
\return HRESULT success or error code.
*/
HRESULT BcmGpioControllerClass::_mapController()
{
	HRESULT hr = S_OK;
	PVOID baseAddress = nullptr;

	hr = GetControllerBaseAddress(
		pi2GpioDeviceName,
		m_hController,
		baseAddress,
		FILE_SHARE_READ | FILE_SHARE_WRITE);

	if (SUCCEEDED(hr))
	{
		m_registers = (PBCM_GPIO)baseAddress;
	}

	return hr;
}

HRESULT BcmGpioControllerClass::mapIfNeeded()
{
	HRESULT hr = S_OK;
	if (m_hController == INVALID_HANDLE_VALUE) hr = _mapController();
	return hr;
}

/**
This method assumes the caller has checked the input parameters.
\param[in] gpioNo The GPIO number of the pad to set. Range: 0-31.
\param[in] state State to set the pad to. 0 - LOW, 1 - HIGH.
\return HRESULT error or success code.
*/
HRESULT BcmGpioControllerClass::setPinState(ULONG gpioNo, ULONG state)
{
	HRESULT hr = mapIfNeeded();
	ULONG bitMask;

	if (SUCCEEDED(hr)) {
		if (gpioNo < 32) {
			bitMask = 1 << gpioNo;
			if (state == 0) m_registers->GPCLR0 = bitMask;
			else m_registers->GPSET0 = bitMask;
		} else {
			bitMask = 1 << (gpioNo - 32);
			if (state == 0) m_registers->GPCLR1 = bitMask;
			else m_registers->GPSET1 = bitMask;
		}
	}

	return hr;
}

/**
This method assumes the caller has checked the input parameters.
\param[in] gpioNo The GPIO number of the pad to read. Range: 0-31.
\param[out] state Set to the state of the input bit.  0 - LOW, 1 - HIGH.
\return HRESULT error or success code.
*/
HRESULT BcmGpioControllerClass::getPinState(ULONG gpioNo, ULONG & state)
{
	HRESULT hr = mapIfNeeded();
	if (SUCCEEDED(hr)) {
		if (gpioNo < 32) state = ((m_registers->GPLEV0) >> gpioNo) & 1;
		else state = ((m_registers->GPLEV1) >> (gpioNo - 32)) & 1;
	}

	return hr;
}

HRESULT BcmGpioControllerClass::attachInterrupt(ULONG intNo, std::function<void(void)> func, ULONG mode)
{
	HRESULT hr = S_OK;

	hr = mapIfNeeded();

	// Tell the driver to attach the interrupt.
	if (SUCCEEDED(hr)){
		hr = m_gpioInterrupts.attachInterrupt(intNo, func, mode, m_hController);
	}

	return hr;
}
#endif // defined(_M_ARM)

#if defined(_M_ARM)
/// Method to detach an interrupt for a GPIO port bit.
HRESULT BcmGpioControllerClass::detachInterrupt(ULONG intNo)
{
	HRESULT hr = S_OK;
	hr = mapIfNeeded();

	// Tell the driver to detach the interrupt.
	if (SUCCEEDED(hr)){
		hr = m_gpioInterrupts.detachInterrupt(intNo, m_hController);
	}

	return hr;
}

/**
This method assumes the caller has checked the input parameters.  This method has
the side effect of setting the pin to GPIO use (since the BCM GPIO controller uses
the same bits to control both function and direction).
\param[in] gpioNo The GPIO number of the pad to configure. Range: 0-53.
\param[in] mode The mode to set for the bit.  Range: DIRECTION_IN or DIRECTION_OUT
\return HRESULT error or success code.
*/
HRESULT BcmGpioControllerClass::setPinDirection(ULONG gpioNo, ULONG mode)
{
	HRESULT hr = S_OK;
	ULONG funcSelData = 0;

	hr = mapIfNeeded();
	if (SUCCEEDED(hr)) {
		funcSelData = m_registers->GPFSELN[gpioNo / 10];   // Read function register data
		funcSelData &= ~(0x07 << ((gpioNo % 10) * 3));	  // Clear bits for GPIO (make input)
		if (mode == DIRECTION_OUT) funcSelData |= (0x01 << ((gpioNo % 10) * 3));   // Set one bit for GPIO (make output)
		m_registers->GPFSELN[gpioNo / 10] = funcSelData;   // Write function register data back
	}

	return hr;
}

/**
This method assumes the caller has checked the input parameters.
\param[in] gpioNo The GPIO number of the pad to configure.  Range: 0-53.
\param[in] function The function to set for the pin.  Range: 0-1.  Function 0 is GPIO, and
function 1 is the non-GPIO use of the pin (called "alternate function 0" in the datasheet).
\return HRESULT error or success code.
*/
HRESULT BcmGpioControllerClass::setPinFunction(ULONG gpioNo, ULONG function)
{
	HRESULT hr = S_OK;
	ULONG funcSelData = 0;

	hr = mapIfNeeded();

	if (SUCCEEDED(hr)) {
		funcSelData = m_registers->GPFSELN[gpioNo / 10];   // Read function register data
		if (((funcSelData & (0x06 << ((gpioNo % 10) * 3))) != 0) || function != 0) {
			funcSelData &= ~(0x07 << ((gpioNo % 10) * 3));	  // Clear bits for (make GPIO input)
			if (function == 1) funcSelData |= (0x04 << ((gpioNo % 10) * 3));   // Set function code to 100b
			m_registers->GPFSELN[gpioNo / 10] = funcSelData;   // Write function register data back
		}
	}

	return hr;
}

/**
This method assumes the caller has checked the input parameters.
\param[in] gpioNo The GPIO number of the pad to configure.  Range: 0-53.
\param[in] pullup TRUE to turn pullup on for this pin, FALSE to turn it off.
\return HRESULT error or success code.
*/
HRESULT BcmGpioControllerClass::setPinPullup(ULONG gpioNo, BOOL pullup)
{
	HRESULT hr = S_OK;
	ULONG gpioPull = 0;
	HiResTimerClass timer;

	hr = mapIfNeeded();
	if (SUCCEEDED(hr)) {
		if (pullup) gpioPull = pullupOn;
		else gpioPull = pullupOff;

		//
		// The sequence to set pullup/down for a pin is:
		// 1) Write desired state to GPPUD
		// 2) Wait 150 cycles
		// 3) Write clock mask to GPPUDCLK0/1 with bits set that correspond to pins to be configured
		// 4) Wait 150 cycles
		// 5) Write to GPPUD to remove state
		// 6) Write to GPPUDCLK0/1 to remove clock bits
		//
		// 150 cycles is 0.25 microseconds with a cpu clock of 600 Mhz.
		//
		m_registers->GPPUD = gpioPull;		 // 1)

		timer.StartTimeout(1);
		while (!timer.TimeIsUp());			  // 2)

		m_registers->GPPUDCLK0 = 1 << gpioNo;
		m_registers->GPPUDCLK1 = 0;			// 3)

		timer.StartTimeout(1);
		while (!timer.TimeIsUp());			  // 4)

		m_registers->GPPUD = 0;				// 5)

		m_registers->GPPUDCLK0 = 0;
		m_registers->GPPUDCLK1 = 0;			// 6)
	}

	return hr;
}

HRESULT BcmGpioControllerClass::setPinMode(ULONG gpioNo, ULONG mode)
{
	HRESULT hr;
	hr = setPinFunction(gpioNo, 0);
	if (FAILED(hr)) goto EXIT;

	hr = setPinDirection(gpioNo, (mode == INPUT || mode == INPUT_PULLUP) ? DIRECTION_IN : DIRECTION_OUT);
	if (FAILED(hr)) goto EXIT;

	hr = setPinPullup(gpioNo, (mode == INPUT_PULLUP));
	if (FAILED(hr)) goto EXIT;

EXIT:
	return hr;
}

#endif // defined(_M_ARM)

